import argparse
import importlib
import inspect
import os
from pathlib import Path

from gradio.blocks import BlockContext
from gradio.components import Component
from tomlkit import dumps, parse

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Description of your program")
    parser.add_argument("-m", "--mode", help="Build mode or dev mode")
    args = parser.parse_args()

    with open("../pyproject.toml") as f:
        pyproject_source = f.read()

    pyproject_toml = parse(pyproject_source)
    keywords = pyproject_toml["project"]["keywords"]
    custom_component = (
        "gradio-custom-component" in keywords or "gradio custom component" in keywords
    )
    if not custom_component:
        exit(0)

    module_name = pyproject_toml["project"]["name"]
    module = importlib.import_module(module_name)

    artifacts: list[str] = pyproject_toml["tool"]["hatch"]["build"]["artifacts"]

    def get_relative_path(path):
        return (
            os.path.abspath(Path(__file__).parent / path)
            .replace(os.path.abspath(os.getcwd()), "")
            .lstrip("/")
        )

    for name in dir(module):
        value = getattr(module, name)
        if name.startswith("__"):
            continue

        if inspect.isclass(value) and (
            issubclass(value, BlockContext) or issubclass(value, Component)
        ):
            file_location = Path(inspect.getfile(value)).parent

            found = [
                x
                for x in artifacts
                if get_relative_path(Path("..") / x)
                == get_relative_path(file_location / value.TEMPLATE_DIR)
            ]
            if len(found) == 0:
                artifacts.append(
                    os.path.abspath(file_location / value.TEMPLATE_DIR)
                    .replace(os.path.abspath(Path("..")), "")
                    .lstrip("/")
                )

            print(
                f"{name}~|~|~|~{os.path.abspath(file_location  / value.TEMPLATE_DIR)}~|~|~|~{os.path.abspath(file_location / value.FRONTEND_DIR)}~|~|~|~{value.get_component_class_id()}"
            )
            continue

    if args.mode == "build":
        pyproject_toml["tool"]["hatch"]["build"]["artifacts"] = artifacts

        with open("../pyproject.toml", "w") as f:
            f.write(dumps(pyproject_toml))



from aiconfig import AIConfigRuntime, ModelParserRegistry
from aiconfig_extension_hugging_face import (
    HuggingFaceAutomaticSpeechRecognitionTransformer,
    HuggingFaceImage2TextTransformer,
    HuggingFaceText2ImageDiffusor,
    HuggingFaceText2SpeechTransformer,
    HuggingFaceTextGenerationTransformer,
    HuggingFaceTextSummarizationTransformer,
    HuggingFaceTextTranslationTransformer,
)


# Example of how users can register model parsers for use in the GradioWorkbook
# The implementation looks for a parsers_path (model_parsers.py by default) which
# should include a module with a register_model_parsers function.
# Here we are registering all the local HuggingFace model parsers as an example
def register_model_parsers() -> None:
    """Register model parsers for local HuggingFace models."""
    automatic_speech_recognition = HuggingFaceAutomaticSpeechRecognitionTransformer()
    AIConfigRuntime.register_model_parser(
        automatic_speech_recognition, "Automatic Speech Recognition (Local)"
    )

    image_to_text = HuggingFaceImage2TextTransformer()
    AIConfigRuntime.register_model_parser(image_to_text, "Image-to-Text (Local)")

    text_to_image = HuggingFaceText2ImageDiffusor()
    AIConfigRuntime.register_model_parser(text_to_image, "Text-to-Image (Local)")

    text_to_speech = HuggingFaceText2SpeechTransformer()
    AIConfigRuntime.register_model_parser(text_to_speech, "Text-to-Speech (Local)")

    text_generation = HuggingFaceTextGenerationTransformer()
    AIConfigRuntime.register_model_parser(text_generation, "Text Generation (Local)")

    text_summarization = HuggingFaceTextSummarizationTransformer()
    AIConfigRuntime.register_model_parser(text_summarization, "Summarization (Local)")

    text_translation = HuggingFaceTextTranslationTransformer()
    AIConfigRuntime.register_model_parser(text_translation, "Translation (Local)")

    # By default, model parsers will also have their own ids registered. Remove those
    # since we just want the task-based names registered
    parsers = [
        automatic_speech_recognition,
        image_to_text,
        text_to_image,
        text_to_speech,
        text_generation,
        text_summarization,
        text_translation,
    ]

    for parser in parsers:
        ModelParserRegistry.remove_model_parser(parser.id())



"""Helper class to reference the AIConfigRuntime state
"""
import copy
import heapq
import time
from threading import Event
from typing import Any, Dict, Literal

from aiconfig import AIConfigRuntime, ModelParserRegistry
from aiconfig.registry import update_model_parser_registry_with_config_runtime
from aiconfig_extension_hugging_face import (
    HuggingFaceAutomaticSpeechRecognitionRemoteInference,
    HuggingFaceConversationalRemoteInference,
    HuggingFaceImage2TextRemoteInference,
    HuggingFaceText2ImageRemoteInference,
    HuggingFaceText2SpeechRemoteInference,
    HuggingFaceTextGenerationRemoteInference,
    HuggingFaceTextSummarizationRemoteInference,
    HuggingFaceTextTranslationRemoteInference,
)

from .session_data import SessionData
from .utils import (
    EXCLUDE_OPTIONS,
    get_validated_path,
    load_user_parser_module,
    show_debug,
)

# TODO (rossdanlm): Use os.path to get better relative path to file
DEFAULT_CONFIG_PATH: Literal = "my_app.aiconfig.json"
DEFAULT_PARSERS_PATH: Literal = "model_parsers.py"
DEFAULT_AICONFIG_SETTINGS: Dict[str, Any] = {
    "name": "Gradio Notebook AIConfig",
    # TODO (rossdanlm): Link this description to our README docs to help people get
    # started
    "description": "This is the AIConfig that is used for the current Gradio notebook",
}
NUM_MINUTES_BEFORE_DELETING_OLD_SESSIONS = 120  # 2 hours


class AIConfigManager:
    """
    Manages the mapping of client session --> SessionData state so that
    we can reference AIConfig from other classes without worrying about it
    being stale. This also ensures that there are no circular dependencies for
    classes that need to reference the AIConfigRuntime

    Also will contain utility methods if needed
    """

    session_data_map: Dict[str, SessionData]
    session_id_lru_min_heap: list[(float, str)]  # (update_time, session_id)
    thread_events: dict[str, Event]

    def __init__(self, config_path: str, parsers_path: str):
        self._clear_default_model_parsers()
        self._register_model_parsers(parsers_path)
        self.session_data_map = {
            "original": SessionData(
                config=self._create_or_load_aiconfig(config_path),
                update_time=time.time(),
            )
        }
        self.session_id_lru_min_heap = []  # Do not store original in min_heap
        self.thread_events = {}

    def _clear_default_model_parsers(self):
        """
        By default, there are a ton of non-hf models/parsers registered in the
        ModelParserRegistry. We want to clear these out so that we can register
        only the hf ones to start
        """
        ModelParserRegistry.clear_registry()

    def _register_model_parsers(self, parsers_path: str):
        """
        Register the model parsers to use for the AIConfig.
        By default, we register the main HuggingFace parsers.

        TODO: Support user-provider parser registration
        """
        automatic_speech_recognition = (
            HuggingFaceAutomaticSpeechRecognitionRemoteInference()
        )
        AIConfigRuntime.register_model_parser(
            automatic_speech_recognition, "Automatic Speech Recognition"
        )

        conversational = HuggingFaceConversationalRemoteInference()
        AIConfigRuntime.register_model_parser(conversational, "Conversational")

        image_to_text = HuggingFaceImage2TextRemoteInference()
        AIConfigRuntime.register_model_parser(image_to_text, "Image-to-Text")

        text_to_image = HuggingFaceText2ImageRemoteInference()
        AIConfigRuntime.register_model_parser(text_to_image, "Text-to-Image")

        text_to_speech = HuggingFaceText2SpeechRemoteInference()
        AIConfigRuntime.register_model_parser(text_to_speech, "Text-to-Speech")

        text_generation = HuggingFaceTextGenerationRemoteInference()
        AIConfigRuntime.register_model_parser(text_generation, "Text Generation")

        text_summarization = HuggingFaceTextSummarizationRemoteInference()
        AIConfigRuntime.register_model_parser(text_summarization, "Summarization")

        text_translation = HuggingFaceTextTranslationRemoteInference()
        AIConfigRuntime.register_model_parser(text_translation, "Translation")

        # By default, model parsers will also have their own ids registered. Remove those
        # since we just want the task-based names registered
        parsers = [
            automatic_speech_recognition,
            conversational,
            image_to_text,
            text_to_image,
            text_to_speech,
            text_generation,
            text_summarization,
            text_translation,
        ]

        for parser in parsers:
            ModelParserRegistry.remove_model_parser(parser.id())

        # Lastly, register any user-provided model parsers, if applicable
        if not parsers_path:
            print(
                f"Warning, no parsers_path was provided so using default path '{DEFAULT_PARSERS_PATH}' instead"
            )
            parsers_path = DEFAULT_PARSERS_PATH

        self._load_user_parser_module_if_exists(parsers_path)

    def _load_user_parser_module_if_exists(self, parsers_module_path: str) -> None:
        try:
            parsers_path = get_validated_path(parsers_module_path)
            load_user_parser_module(parsers_path)
            print(f"Loaded parsers module from {parsers_module_path}")
        except Exception as e:
            print(f"Failed to load parsers module: {e}")

    def delete_session_id(self, session_id: str) -> None:
        """Delete the session_id from the session_data_map map"""
        self.session_data_map.pop(session_id, None)

    def get_config(self, session_id: str) -> AIConfigRuntime:
        """Get a config that is mapped to a session id"""
        update_time = time.time()
        if session_id not in self.session_data_map:
            copied_config = copy.deepcopy(self.session_data_map["original"].config)
            session_data = SessionData(config=copied_config, update_time=update_time)
            self.session_data_map[session_id] = session_data
            heapq.heappush(self.session_id_lru_min_heap, (update_time, session_id))

        if show_debug():
            print(f"{self.session_data_map.keys()=}")
            update_times = [
                (k, v.update_time) for k, v in self.session_data_map.items()
            ]
            print(f"{update_times=}")

        self.session_data_map[session_id].update_time = update_time
        self.clear_old_session_ids()
        return self.session_data_map[session_id].config

    def get_config_json(self, session_id: str) -> dict[str, Any]:
        """Helper function to return the config in json dict format"""
        return self.get_config(session_id).model_dump(exclude=EXCLUDE_OPTIONS)

    def set_config(self, session_id: str, config: AIConfigRuntime):
        """Set the AIConfigRuntime for a session_id and also reset the update_time"""
        session_data = SessionData(config=config, update_time=time.time())
        self.session_data_map[session_id] = session_data

    def _create_or_load_aiconfig(self, config_path: str) -> AIConfigRuntime:
        """
        Create or load an AIConfigRuntime from a provide config_path.
        This should only ever be called during init of this AIConfigManager class
        """
        already_tried_default_filepath = False
        if not config_path:
            print(
                f"Warning, no config_path was provided so using default path '{DEFAULT_CONFIG_PATH}' instead"
            )
            config_path = DEFAULT_CONFIG_PATH
            already_tried_default_filepath = True

        try:
            config = AIConfigRuntime.load(config_path)
        # TODO (rossdanlm): Test this also with malformed json format to see which error it produces and catch for that
        except FileNotFoundError:
            try:
                if not already_tried_default_filepath:
                    print(
                        f"Warning, config_path '{config_path}' not found, trying default config_path '{DEFAULT_CONFIG_PATH}' instead..."
                    )
                    config_path = DEFAULT_CONFIG_PATH
                    config = AIConfigRuntime.load(config_path)
                else:
                    raise FileNotFoundError()
            except FileNotFoundError:
                print(
                    f"Warning, config_path '{config_path}' not found, creating new AIConfig."
                )
                config_path = DEFAULT_CONFIG_PATH
                config = AIConfigRuntime.create(**DEFAULT_AICONFIG_SETTINGS)
        config.file_path = config_path
        update_model_parser_registry_with_config_runtime(config)
        return config

    def get_models(self) -> list[str]:
        """Helper function to get the models from the ModelParserRegistry"""
        return ModelParserRegistry.parser_ids()

    def clear_old_session_ids(self):
        """
        Clear out old session_ids from the session_id_lru_min_heap and session_data_map
        """
        threshold_cutoff_time: float = (
            time.time() - NUM_MINUTES_BEFORE_DELETING_OLD_SESSIONS * 60
        )
        if show_debug():
            print(f"{threshold_cutoff_time=}\n")
        while (
            len(self.session_id_lru_min_heap) > 0
            and self.session_id_lru_min_heap[0][0]  # update_time
            < threshold_cutoff_time
        ):
            session_id = self.session_id_lru_min_heap[0][1]
            should_delete = self.remove_lru_session_if_not_updated(
                threshold_cutoff_time,
            )
            if should_delete:
                self.session_data_map.pop(session_id, None)

    def remove_lru_session_if_not_updated(self, threshold_cutoff_time: float) -> bool:
        """
        Remove the least recently used session from the session_id_lru_min_heap
        If the session has been updated since it was added to the min_heap, and the
        update time is after the threshold cutoff, then we need to add it back to the
        min_heap and check again

        @return bool: Whether the session was removed or not
        """
        old_update_time, session_id = heapq.heappop(self.session_id_lru_min_heap)
        actual_update_time = self.session_data_map[session_id].update_time
        if (
            old_update_time < actual_update_time
            and actual_update_time > threshold_cutoff_time
        ):
            # This means that the update_time was updated since we added it to the
            # min_heap, so we need to add it back to the min_heap and check again
            heapq.heappush(
                self.session_id_lru_min_heap, (actual_update_time, session_id)
            )
            return False
        return True



""""
File for defining the event data classes: https://www.gradio.app/docs/eventdata.
In general, the value will be the payload passed from the frontend AIConfig as 
a json string
"""
import asyncio
import copy
import ctypes
import json
import random
import threading
import uuid
from datetime import datetime
from typing import TYPE_CHECKING, Any, Dict, Optional, Union

import boto3
import requests
from aiconfig import AIConfigRuntime, InferenceOptions
from aiconfig.schema import ExecuteResult, Output, Prompt
from botocore import UNSIGNED
from botocore.config import Config
from gradio.events import EventData

from .aiconfig_manager import AIConfigManager
from .utils import (
    EXCLUDE_OPTIONS,
    LASTMILE_BASE_URI,
    STOP_STREAMING_SIGNAL,
    QueueIterator,
    show_debug,
)

if TYPE_CHECKING:
    from gradio.blocks import Block


class EventWithSessionIdData(EventData):
    """
    In order to prevent multiple clients from overwriting the same AIConfig
    in a given HuggingFace workspace, we need to manually pass in the
    `session_id` from the Gradio client for all events.
    """

    def __init__(self, target: Union["Block", None], data: Any):
        super().__init__(target, data)

        # Session ID for the current session
        self.session_id: str = data["session_id"]


class AddPromptEventData(EventWithSessionIdData):
    """
    Add a prompt to the component's AIConfig.

        Default position is at the end for now
        TODO (rossdanlm): Add index to specify what position we should add
        prompt to AIConfig
    """

    def __init__(self, target: Union["Block", None], data: Any):
        super().__init__(target, data)

        self.prompt_name: str = data["prompt_name"]

        # Data of the Prompt to be added
        self.prompt: Prompt = data["prompt"]

        # Index to add the prompt to
        self.index: int = data["index"]


class CancelRunEventData(EventData):
    """
    Cancel the prompt run operation identified by the cancellation_token field
    """

    def __init__(self, target: Union["Block", None], data: Any):
        super().__init__(target, data)

        # Token generated and stored by prompt client whenever the run prompt
        # button is clicked to start running the prompt
        self.cancellation_token_id: str = data["cancellation_token_id"]


class DeletePromptEventData(EventWithSessionIdData):
    """
    Delete the identified by the prompt_name field
    """

    def __init__(self, target: Union["Block", None], data: Any):
        super().__init__(target, data)

        # Name of the prompt to be deleted
        self.prompt_name: str = data["prompt_name"]


class RunPromptEventData(EventWithSessionIdData):
    """
    Run the prompt identified by the prompt_name field
    """

    def __init__(self, target: Union["Block", None], data: Any):
        super().__init__(target, data)

        # Name of the prompt to be run
        self.prompt_name: str = data["prompt_name"]

        # Token to use for cancelling this particular run
        self.cancellation_token: Optional[str] = data.get("cancellation_token", None)

        # Whether we should run the other prompts that this depends on
        # TODO(rossdanlm): Make this generic kwargs
        self.run_with_dependencies: bool = data.get("run_with_dependencies", True)

        self.api_token: Optional[str] = data.get("api_token", None)

        # TODO(rossdanlm): Add inference options (streaming) here?


class SetConfigDescriptionEventData(EventWithSessionIdData):
    """
    Set the description for the AIConfig
    """

    def __init__(self, target: Union["Block", None], data: Any):
        super().__init__(target, data)

        # Description to set for the AIConfig
        self.description: str = data["description"]


class SetConfigNameEventData(EventWithSessionIdData):
    """
    Set the name for the AIConfig
    """

    def __init__(self, target: Union["Block", None], data: Any):
        super().__init__(target, data)

        # Name to set for the AIConfig
        self.name: str = data["name"]


class SetParametersEventData(EventWithSessionIdData):
    """
    Set parameter values either for the overall config or a specific prompt
    identified by the prompt_name field
    """

    def __init__(self, target: Union["Block", None], data: Any):
        super().__init__(target, data)

        # Parameter name key to be set
        self.parameters: Dict[str, Any] = data["parameters"]

        # Name of the prompt to be updated. If None, update the overall config
        self.prompt_name: str = data.get("prompt_name", None)


class ShareConfigEventData(EventWithSessionIdData):
    """
    Create a shareable link for the AIConfig that will display a read-only mode of the
    AIConfigEditor after loading the AIConfig json to S3 and then passing the S3 url and
    HuggingFace workspace url to LastMileAI api/aiconfig/upload endpoint
    """

    def __init__(self, target: Union["Block", None], data: Any):
        super().__init__(target, data)

        # The HuggingFace (or localhost) URL where this Gradio Workbook is hosted
        self.workspace_url: str = data.get("workspace_url", None)


class UpdateModelEventData(EventWithSessionIdData):
    """
    Update the model either for the overall config or a specific prompt
    identified by the prompt_name field
    """

    def __init__(self, target: Union["Block", None], data: Any):
        super().__init__(target, data)

        # Model name to update the prompt (if not None) or AIConfig with
        self.model_name: Optional[str] = data.get("model_name", None)

        # Model settings to update the prompt (if not None) or AIConfig with
        self.model_settings: Optional[Dict[str, Any]] = data.get("model_settings", None)

        # Name of the prompt to be updated. If None, update the overall config
        self.prompt_name: Optional[str] = data.get("prompt_name", None)


class UpdatePromptEventData(EventWithSessionIdData):
    """
    Update the prompt identified by the prompt_name field
    """

    def __init__(self, target: Union["Block", None], data: Any):
        super().__init__(target, data)

        # Name of the prompt to be updated
        self.prompt_name: str = data["prompt_name"]

        # Data to update the prompt with
        self.prompt: Prompt = data["prompt"]


class EventHandler:
    """
    A class to store the event implementation for a specific event type
    """

    config_manager: AIConfigManager

    def __init__(self, config_manager: AIConfigManager):
        self.config_manager = config_manager

    def add_prompt_impl(self, event: AddPromptEventData) -> str:
        """
        Run this for the add_prompt event
        """
        session_id = event.session_id
        prompt_name = event.prompt_name
        prompt_json_str = json.dumps(event.prompt)
        index = event.index

        prompt = Prompt.model_validate_json(prompt_json_str)

        config: AIConfigRuntime = self.config_manager.get_config(session_id)
        config.add_prompt(prompt_name=prompt_name, prompt_data=prompt, index=index)
        return json.dumps({"aiconfig": self.config_manager.get_config_json(session_id)})

    def cancel_run_impl(self, event: CancelRunEventData) -> str:
        """
        Run this for the cancel_run event
        """
        cancellation_token_id: str = event.cancellation_token_id
        if cancellation_token_id is not None:
            thread_event = self.config_manager.thread_events.get(cancellation_token_id)
            if thread_event is not None:
                thread_event.set()
                self.config_manager.thread_events.pop(cancellation_token_id)
                return json.dumps({"cancellation_token_id": cancellation_token_id})

            # Return a 422 Unprocessable Entity
            error_info = {
                "error": {
                    "cancellation_token_id": cancellation_token_id,
                    "message": "Unable to process cancellation request. Task not found for associated cancellation_token_id",
                    "code": 422,
                }
            }
            return json.dumps(error_info)

        # Return a 400 Bad Request error
        error_info = {
            "error": {
                "message": "No cancellation_token_id was specified in the request. Unable to process cancellation.",
                "code": 400,
            }
        }
        return json.dumps(error_info)

    def clear_outputs_impl(self, event: EventWithSessionIdData) -> str:
        """
        Run this for the clear_outputs event
        """
        session_id = event.session_id
        config: AIConfigRuntime = self.config_manager.get_config(session_id)
        for prompt in config.prompts:
            prompt_name = prompt.name
            config.delete_output(prompt_name)
        return json.dumps({"aiconfig": self.config_manager.get_config_json(session_id)})

    def delete_prompt_impl(self, event: DeletePromptEventData) -> str:
        """
        Run this for the delete_prompt event. We don't need to return AIConfig
        for the Gradio frontend
        """
        session_id = event.session_id
        config: AIConfigRuntime = self.config_manager.get_config(session_id)
        config.delete_prompt(event.prompt_name)
        return json.dumps({})

    def get_aiconfig_impl(self, event: EventWithSessionIdData) -> str:
        """
        Run this for the get_aiconfig event
        """
        session_id = event.session_id
        return json.dumps({"aiconfig": self.config_manager.get_config_json(session_id)})

    def remove_session_id_impl(self, event: EventWithSessionIdData) -> str:
        """
        Run this for the remove_session_id event
        """
        session_id = event.session_id
        self.config_manager.delete_session_id(session_id)
        return json.dumps({})

    def run_prompt_impl(self, event: RunPromptEventData) -> str:
        """
        Run this for the run_prompt event
        """
        try:
            session_id = event.session_id
            executing_config: AIConfigRuntime = self.config_manager.get_config(
                session_id
            )

            prompt_name = event.prompt_name
            params = executing_config.get_parameters(prompt_name)

            # Define stream callback and queue object for streaming results
            output_text_queue = QueueIterator()

            def update_output_queue(data, _accumulated_data, _index) -> None:
                should_end_stream = data == STOP_STREAMING_SIGNAL
                output_text_queue.put(data, should_end_stream)

            inference_options = InferenceOptions(
                stream=True,  # Stream is always true for Gradio server implementation
                stream_callback=update_output_queue,
                api_token=event.api_token,
            )

            # Deepcopy the aiconfig prior to run so we can restore it in the case
            # the run operation is cancelled or encounters some error
            pre_run_config = copy.deepcopy(executing_config)

            cancellation_token_id = event.cancellation_token
            if not cancellation_token_id:
                cancellation_token_id = str(uuid.uuid4())
            self.config_manager.thread_events[cancellation_token_id] = threading.Event()

            def generate(cancellation_token_id: str):  # type: ignore
                # Use multi-threading so that we don't block run command from
                # displaying the streamed output (if streaming is supported)
                def run_async_config_in_thread():
                    try:
                        asyncio.run(
                            executing_config.run(
                                prompt_name=prompt_name,
                                params=params,
                                run_with_dependencies=False,
                                options=inference_options,
                            )
                        )
                    except Exception as e:
                        output_text_queue.put(e)
                    output_text_queue.put(STOP_STREAMING_SIGNAL)  # type: ignore

                def create_error_payload(message: str, code: int):
                    aiconfig_json = (
                        pre_run_config.model_dump(exclude=EXCLUDE_OPTIONS)
                        if pre_run_config is not None
                        else None
                    )
                    error_info = {
                        "error": {
                            "message": message,
                            "code": code,
                            "data": aiconfig_json,
                        }
                    }
                    return json.dumps(error_info)

                def create_cancellation_payload():
                    return create_error_payload(
                        message="The task was cancelled.", code=499
                    )

                def handle_cancellation():
                    yield create_cancellation_payload()
                    # Reset the aiconfig state to the state prior to the run,
                    # and kill the running thread
                    kill_thread(t.ident)
                    self.config_manager.set_config(session_id, pre_run_config)

                def kill_thread(thread_id: int | None):
                    """
                    Kill the thread with the given thread_id.

                    PyThreadState_SetAsyncExc: This is a C API function in Python
                    which is used to raise an exception in the context of the
                    specified thread.

                    SystemExit: This is the exception we'd like to raise in the
                    target thread.
                    """
                    if thread_id is None:
                        # Nothing to do
                        return
                    response = ctypes.pythonapi.PyThreadState_SetAsyncExc(
                        ctypes.c_long(thread_id), ctypes.py_object(SystemExit)
                    )

                    if response == 0:
                        print(f"Invalid thread id {thread_id}")
                    elif response != 1:
                        # If the response is not 1, the function didn't work
                        # correctly, and you should call it again with
                        # exc=NULL to reset it.
                        ctypes.pythonapi.PyThreadState_SetAsyncExc(thread_id, None)

                cancellation_event = self.config_manager.thread_events[
                    cancellation_token_id
                ]
                t = threading.Thread(target=run_async_config_in_thread)
                t.start()

                # If model supports streaming, need to wait until streamer has at
                # least 1 item to display. If model does not support streaming,
                # need to wait until the aiconfig.run() thread is complete
                while output_text_queue.isEmpty() and t.is_alive():
                    if cancellation_event.is_set():
                        yield from handle_cancellation()
                        return

                if not output_text_queue.isEmpty():
                    accumulated_output_text = ""
                    for text in output_text_queue:
                        if cancellation_event.is_set():
                            yield from handle_cancellation()
                            return

                        if isinstance(text, Exception):
                            yield create_error_payload(
                                message=f"Exception: {text}", code=500
                            )
                            return
                        elif isinstance(text, str):
                            accumulated_output_text += text
                        elif isinstance(text, dict) and "content" in text:
                            # TODO: Fix streaming output format so that it returns text
                            accumulated_output_text += text["content"]
                        elif isinstance(text, dict) and "generated_text" in text:
                            # TODO: Fix streaming output format so that it returns text
                            accumulated_output_text += text["generated_text"]

                        accumulated_output: Output = ExecuteResult(
                            **{
                                "output_type": "execute_result",
                                "data": accumulated_output_text,
                                # Assume streaming only supports single output
                                # I think this actually may be wrong for PaLM or OpenAI
                                # TODO: Need to sync with Ankush but can fix forward
                                "execution_count": 0,
                                "metadata": {},
                            }  # type: ignore
                        )
                        if show_debug():
                            print(f"{accumulated_output_text=}")
                        yield json.dumps({"output_chunk": accumulated_output.to_json()})

                # Ensure that the run process is complete to yield final output
                t.join()

                if cancellation_event.is_set():
                    yield from handle_cancellation()
                    return

                self.config_manager.thread_events.pop(cancellation_token_id, None)
                aiconfig_json = self.config_manager.get_config_json(session_id)
                yield json.dumps({"aiconfig_chunk": aiconfig_json})
                yield json.dumps({"stop_streaming": True})

            if show_debug():
                print(f"Running `aiconfig.run()` command with request: {event}")
            yield from generate(cancellation_token_id)
        except Exception as e:
            # Return a 400 Bad Request error
            error_info = {
                "error": {
                    "message": f"Failed to run prompt: {type(e)}, {e}",
                    "code": 400,
                    "data": self.config_manager.get_config_json(session_id),
                }
            }
            yield json.dumps(error_info)

    def set_config_description_impl(self, event: SetConfigDescriptionEventData) -> str:
        """
        Run this for the set_config_description event. We don't need to return AIConfig
        for the Gradio frontend
        """
        session_id = event.session_id
        config: AIConfigRuntime = self.config_manager.get_config(session_id)
        config.set_description(event.description)
        return json.dumps({})

    def set_config_name_impl(self, event: SetConfigNameEventData) -> str:
        """
        Run this for the set_config_name event. We don't need to return AIConfig
        for the Gradio frontend
        """
        session_id = event.session_id
        config: AIConfigRuntime = self.config_manager.get_config(session_id)
        config.set_name(event.name)
        return json.dumps({})

    def set_parameters_impl(self, event: SetParametersEventData) -> str:
        """
        Run this for the set_parameters event. We don't need to return AIConfig
        for the Gradio frontend
        """
        session_id = event.session_id
        config: AIConfigRuntime = self.config_manager.get_config(session_id)
        config.set_parameters(event.parameters, event.prompt_name)
        return json.dumps({})

    def share_config_impl(self, event: ShareConfigEventData) -> str:
        """
        Run this for the share_config event. See documentation on how to upload to S3:
        https://boto3.amazonaws.com/v1/documentation/api/latest/guide/s3-uploading-files.html
        """
        session_id = event.session_id
        workspace_url = event.workspace_url
        config: AIConfigRuntime = self.config_manager.get_config(session_id)
        # s3 file uris cannot have '+' character, so replace with '_'
        filename: str = f"{_sanitize_filename(config.name)}.aiconfig.json"

        #   TODO: Add back once CORS is resolved
        #   const policyResponse = await fetch(
        #   "https://lastmileai.dev/api/upload/publicpolicy"
        #   );
        #   const policy = await policyResponse.json();
        random_int: int = random.randint(0, 10001)
        bucket: str = "lastmileai.aiconfig.public"
        upload_key: str = f"aiconfigs/{_get_date_time_string()}/{random_int}/{filename}"

        config_str: str = json.dumps(
            self.config_manager.get_config_json(session_id), indent=2
        )
        config_bytes: bytes = config_str.encode("utf-8")

        s3_client = boto3.client("s3", config=Config(signature_version=UNSIGNED))
        s3_client.put_object(
            Body=config_bytes,
            Bucket=bucket,
            Key=upload_key,
            ACL="public-read",
            ContentType="application/json",
        )

        # For some reason, the URL returned by boto3 doesn't work if you
        # replace " " with "%20" directly within the "put_object" API for
        # the Key field, so instead we are replacing it over here
        s3_url: str = (
            f"https://s3.amazonaws.com/{bucket}/{upload_key.replace(' ', '%20')}"
        )

        lastmile_upload_url: str = LASTMILE_BASE_URI + "api/aiconfig/upload"
        payload = {"url": s3_url, "source": "gradio"}
        if workspace_url is not None:
            payload["sourceUrl"] = workspace_url
        response: requests.Response = requests.post(
            lastmile_upload_url, data=payload, timeout=20
        )
        if response.status_code >= 400:
            response.raise_for_status()

        uploaded_config_id: str = response.json()["id"]
        rendering_uri: str = LASTMILE_BASE_URI + f"aiconfig/{uploaded_config_id}"
        return json.dumps({"share_url": rendering_uri})

    def update_model_impl(self, event: UpdateModelEventData) -> str:
        """
        Run this for the update_model event
        """
        session_id = event.session_id
        config: AIConfigRuntime = self.config_manager.get_config(session_id)
        config.update_model(event.model_name, event.model_settings, event.prompt_name)
        return json.dumps({"aiconfig": self.config_manager.get_config_json(session_id)})

    def update_prompt_impl(self, event: UpdatePromptEventData) -> str:
        """
        Run this for the update_prompt event
        """
        session_id = event.session_id
        prompt_json_str = json.dumps(event.prompt)
        prompt = Prompt.model_validate_json(prompt_json_str)

        config: AIConfigRuntime = self.config_manager.get_config(session_id)
        config.update_prompt(event.prompt_name, prompt)
        return json.dumps({"aiconfig": self.config_manager.get_config_json(session_id)})


def _sanitize_filename(filename: str) -> str:
    """
    Sanitize the filename to remove any characters that are not allowed in
    S3 filenames
    """
    return filename.replace("+", "_")


def _get_date_time_string() -> str:
    now = datetime.now()
    return now.strftime("%Y_%m_%d_%H_%M_%S")



from .aiconfig_manager import AIConfigManager
from .events import (
    AddPromptEventData,
    DeletePromptEventData,
    EventHandler,
    RunPromptEventData,
    UpdatePromptEventData,
)
from .gradio_notebook import GradioNotebook
from .gradio_notebook_component import GradioNotebookComponent
from .utils import STOP_STREAMING_SIGNAL, QueueIterator, show_debug

event_data_classes = [
    "AddPromptEventData",
    "DeletePromptEventData",
    "RunPromptEventData",
    "UpdatePromptEventData",
]
__all__ = event_data_classes + [
    "AIConfigManager",
    "EventHandler",
    "GradioNotebookComponent",
    "GradioNotebook",
    "QueueIterator",
    "STOP_STREAMING_SIGNAL",
    "show_debug",
]



from gradio.components.base import Component
from gradio.data_classes import GradioModel
from gradio.events import EventListener

from .utils import show_debug


class GradioAIConfig(GradioModel):
    # TODO(rossdanlm): Explicitly define AIConfigRuntime
    # https://github.com/lastmile-ai/gradio-workbook/issues/62
    pass

from gradio.events import Dependency

class GradioNotebookComponent(Component):
    data_model = GradioAIConfig

    add_prompt_event_listener = EventListener(
        "add_prompt",
        doc="Triggered when user clicks the Add Prompt button",
    )
    cancel_run_event_listener = EventListener(
        "cancel_run",
        doc="Triggered when user clicks the cancel button while a prompt is running",
    )
    clear_outputs_event_listener = EventListener(
        "clear_outputs",
        doc="Triggered when user clicks the Clear Outputs button",
    )
    delete_prompt_event_listener = EventListener(
        "delete_prompt",
        doc="Triggered when user clicks the Delete Prompt button",
    )
    get_aiconfig_event_listener = EventListener(
        "get_aiconfig",
        doc="Triggered when user needs a non-stale version of config from server to perform an action directly on client (ex: download)",
    )
    remove_session_id_event_listener = EventListener(
        "remove_session_id",
        doc="Triggered when window is closing and session is about to be deleted",
    )
    share_config_event_listener = EventListener(
        "share_config",
        doc="Triggered when user clicks the Share button",
    )
    run_prompt_event_listener = EventListener(
        "run_prompt",
        doc="Triggered when user clicks the Run Prompt button",
    )
    set_config_description_event_listener = EventListener(
        "set_config_description",
        doc="Triggered when user sets the description for the AIConfig",
    )
    set_config_name_event_listener = EventListener(
        "set_config_name",
        doc="Triggered when user sets the name for the AIConfig",
    )
    set_parameters_event_listener = EventListener(
        "set_parameters",
        doc="Triggered when user sets global or prompt parameters. If done with a prompt, \
            it will update the param for that prompt. Otherwise, it will update \
            the overall parameters for the AIConfig",
    )
    update_model_event_listener = EventListener(
        "update_model",
        doc="Triggered when user clicks the Update Model button. If done with a prompt, \
            it will update the model settings for that prompt. Otherwise, it will update \
            the overall model settings for the AIConfig",
    )
    update_prompt_event_listener = EventListener(
        "update_prompt",
        doc="Triggered when user updates the prompt name or settings",
    )
    EVENTS = [
        add_prompt_event_listener,
        cancel_run_event_listener,
        clear_outputs_event_listener,
        delete_prompt_event_listener,
        get_aiconfig_event_listener,
        remove_session_id_event_listener,
        run_prompt_event_listener,
        share_config_event_listener,
        set_config_description_event_listener,
        set_config_name_event_listener,
        set_parameters_event_listener,
        update_model_event_listener,
        update_prompt_event_listener,
    ]

    def preprocess(self, payload: GradioAIConfig | None):
        if show_debug():
            print("\nin preprocess")
            print(f"{payload=}")
        if payload is None:
            return payload
        return payload

    def postprocess(self, value: str) -> str:
        """
        Return an output in the form of a json-formatted string for
        client to process
        """
        if show_debug():
            print("\nin postprocess")
            print(f"{value=}")
        return value

    # TODO (rossdanlm): Update this with a valud AIConfigRuntime object
    # https://github.com/lastmile-ai/gradio-workbook/issues/62
    def example_inputs(self):
        if show_debug():
            print("\nin example_inputs")
        return {"foo": "bar"}

    def api_info(self):
        return {"type": {}, "description": "AIConfigRuntime"}

    
    def add_prompt(self,
        fn: Callable | None,
        inputs: Component | Sequence[Component] | set[Component] | None = None,
        outputs: Component | Sequence[Component] | None = None,
        api_name: str | None | Literal[False] = None,
        scroll_to_output: bool = False,
        show_progress: Literal["full", "minimal", "hidden"] = "full",
        queue: bool | None = None,
        batch: bool = False,
        max_batch_size: int = 4,
        preprocess: bool = True,
        postprocess: bool = True,
        cancels: dict[str, Any] | list[dict[str, Any]] | None = None,
        every: float | None = None,
        trigger_mode: Literal["once", "multiple", "always_last"] | None = None,
        js: str | None = None,
        concurrency_limit: int | None | Literal["default"] = "default",
        concurrency_id: str | None = None,
        show_api: bool = True) -> Dependency:
        """
        Parameters:
            fn: the function to call when this event is triggered. Often a machine learning model's prediction function. Each parameter of the function corresponds to one input component, and the function should return a single value or a tuple of values, with each element in the tuple corresponding to one output component.
            inputs: List of gradio.components to use as inputs. If the function takes no inputs, this should be an empty list.
            outputs: List of gradio.components to use as outputs. If the function returns no outputs, this should be an empty list.
            api_name: Defines how the endpoint appears in the API docs. Can be a string, None, or False. If False, the endpoint will not be exposed in the api docs. If set to None, the endpoint will be exposed in the api docs as an unnamed endpoint, although this behavior will be changed in Gradio 4.0. If set to a string, the endpoint will be exposed in the api docs with the given name.
            scroll_to_output: If True, will scroll to output component on completion
            show_progress: If True, will show progress animation while pending
            queue: If True, will place the request on the queue, if the queue has been enabled. If False, will not put this event on the queue, even if the queue has been enabled. If None, will use the queue setting of the gradio app.
            batch: If True, then the function should process a batch of inputs, meaning that it should accept a list of input values for each parameter. The lists should be of equal length (and be up to length `max_batch_size`). The function is then *required* to return a tuple of lists (even if there is only 1 output component), with each list in the tuple corresponding to one output component.
            max_batch_size: Maximum number of inputs to batch together if this is called from the queue (only relevant if batch=True)
            preprocess: If False, will not run preprocessing of component data before running 'fn' (e.g. leaving it as a base64 string if this method is called with the `Image` component).
            postprocess: If False, will not run postprocessing of component data before returning 'fn' output to the browser.
            cancels: A list of other events to cancel when this listener is triggered. For example, setting cancels=[click_event] will cancel the click_event, where click_event is the return value of another components .click method. Functions that have not yet run (or generators that are iterating) will be cancelled, but functions that are currently running will be allowed to finish.
            every: Run this event 'every' number of seconds while the client connection is open. Interpreted in seconds.
            trigger_mode: If "once" (default for all events except `.change()`) would not allow any submissions while an event is pending. If set to "multiple", unlimited submissions are allowed while pending, and "always_last" (default for `.change()` event) would allow a second submission after the pending event is complete.
            js: Optional frontend js method to run before running 'fn'. Input arguments for js method are values of 'inputs' and 'outputs', return should be a list of values for output components.
            concurrency_limit: If set, this is the maximum number of this event that can be running simultaneously. Can be set to None to mean no concurrency_limit (any number of this event can be running simultaneously). Set to "default" to use the default concurrency limit (defined by the `default_concurrency_limit` parameter in `Blocks.queue()`, which itself is 1 by default).
            concurrency_id: If set, this is the id of the concurrency group. Events with the same concurrency_id will be limited by the lowest set concurrency_limit.
            show_api: whether to show this event in the "view API" page of the Gradio app, or in the ".view_api()" method of the Gradio clients. Unlike setting api_name to False, setting show_api to False will still allow downstream apps to use this event. If fn is None, show_api will automatically be set to False.
        """
        ...
    
    def cancel_run(self,
        fn: Callable | None,
        inputs: Component | Sequence[Component] | set[Component] | None = None,
        outputs: Component | Sequence[Component] | None = None,
        api_name: str | None | Literal[False] = None,
        scroll_to_output: bool = False,
        show_progress: Literal["full", "minimal", "hidden"] = "full",
        queue: bool | None = None,
        batch: bool = False,
        max_batch_size: int = 4,
        preprocess: bool = True,
        postprocess: bool = True,
        cancels: dict[str, Any] | list[dict[str, Any]] | None = None,
        every: float | None = None,
        trigger_mode: Literal["once", "multiple", "always_last"] | None = None,
        js: str | None = None,
        concurrency_limit: int | None | Literal["default"] = "default",
        concurrency_id: str | None = None,
        show_api: bool = True) -> Dependency:
        """
        Parameters:
            fn: the function to call when this event is triggered. Often a machine learning model's prediction function. Each parameter of the function corresponds to one input component, and the function should return a single value or a tuple of values, with each element in the tuple corresponding to one output component.
            inputs: List of gradio.components to use as inputs. If the function takes no inputs, this should be an empty list.
            outputs: List of gradio.components to use as outputs. If the function returns no outputs, this should be an empty list.
            api_name: Defines how the endpoint appears in the API docs. Can be a string, None, or False. If False, the endpoint will not be exposed in the api docs. If set to None, the endpoint will be exposed in the api docs as an unnamed endpoint, although this behavior will be changed in Gradio 4.0. If set to a string, the endpoint will be exposed in the api docs with the given name.
            scroll_to_output: If True, will scroll to output component on completion
            show_progress: If True, will show progress animation while pending
            queue: If True, will place the request on the queue, if the queue has been enabled. If False, will not put this event on the queue, even if the queue has been enabled. If None, will use the queue setting of the gradio app.
            batch: If True, then the function should process a batch of inputs, meaning that it should accept a list of input values for each parameter. The lists should be of equal length (and be up to length `max_batch_size`). The function is then *required* to return a tuple of lists (even if there is only 1 output component), with each list in the tuple corresponding to one output component.
            max_batch_size: Maximum number of inputs to batch together if this is called from the queue (only relevant if batch=True)
            preprocess: If False, will not run preprocessing of component data before running 'fn' (e.g. leaving it as a base64 string if this method is called with the `Image` component).
            postprocess: If False, will not run postprocessing of component data before returning 'fn' output to the browser.
            cancels: A list of other events to cancel when this listener is triggered. For example, setting cancels=[click_event] will cancel the click_event, where click_event is the return value of another components .click method. Functions that have not yet run (or generators that are iterating) will be cancelled, but functions that are currently running will be allowed to finish.
            every: Run this event 'every' number of seconds while the client connection is open. Interpreted in seconds.
            trigger_mode: If "once" (default for all events except `.change()`) would not allow any submissions while an event is pending. If set to "multiple", unlimited submissions are allowed while pending, and "always_last" (default for `.change()` event) would allow a second submission after the pending event is complete.
            js: Optional frontend js method to run before running 'fn'. Input arguments for js method are values of 'inputs' and 'outputs', return should be a list of values for output components.
            concurrency_limit: If set, this is the maximum number of this event that can be running simultaneously. Can be set to None to mean no concurrency_limit (any number of this event can be running simultaneously). Set to "default" to use the default concurrency limit (defined by the `default_concurrency_limit` parameter in `Blocks.queue()`, which itself is 1 by default).
            concurrency_id: If set, this is the id of the concurrency group. Events with the same concurrency_id will be limited by the lowest set concurrency_limit.
            show_api: whether to show this event in the "view API" page of the Gradio app, or in the ".view_api()" method of the Gradio clients. Unlike setting api_name to False, setting show_api to False will still allow downstream apps to use this event. If fn is None, show_api will automatically be set to False.
        """
        ...
    
    def clear_outputs(self,
        fn: Callable | None,
        inputs: Component | Sequence[Component] | set[Component] | None = None,
        outputs: Component | Sequence[Component] | None = None,
        api_name: str | None | Literal[False] = None,
        scroll_to_output: bool = False,
        show_progress: Literal["full", "minimal", "hidden"] = "full",
        queue: bool | None = None,
        batch: bool = False,
        max_batch_size: int = 4,
        preprocess: bool = True,
        postprocess: bool = True,
        cancels: dict[str, Any] | list[dict[str, Any]] | None = None,
        every: float | None = None,
        trigger_mode: Literal["once", "multiple", "always_last"] | None = None,
        js: str | None = None,
        concurrency_limit: int | None | Literal["default"] = "default",
        concurrency_id: str | None = None,
        show_api: bool = True) -> Dependency:
        """
        Parameters:
            fn: the function to call when this event is triggered. Often a machine learning model's prediction function. Each parameter of the function corresponds to one input component, and the function should return a single value or a tuple of values, with each element in the tuple corresponding to one output component.
            inputs: List of gradio.components to use as inputs. If the function takes no inputs, this should be an empty list.
            outputs: List of gradio.components to use as outputs. If the function returns no outputs, this should be an empty list.
            api_name: Defines how the endpoint appears in the API docs. Can be a string, None, or False. If False, the endpoint will not be exposed in the api docs. If set to None, the endpoint will be exposed in the api docs as an unnamed endpoint, although this behavior will be changed in Gradio 4.0. If set to a string, the endpoint will be exposed in the api docs with the given name.
            scroll_to_output: If True, will scroll to output component on completion
            show_progress: If True, will show progress animation while pending
            queue: If True, will place the request on the queue, if the queue has been enabled. If False, will not put this event on the queue, even if the queue has been enabled. If None, will use the queue setting of the gradio app.
            batch: If True, then the function should process a batch of inputs, meaning that it should accept a list of input values for each parameter. The lists should be of equal length (and be up to length `max_batch_size`). The function is then *required* to return a tuple of lists (even if there is only 1 output component), with each list in the tuple corresponding to one output component.
            max_batch_size: Maximum number of inputs to batch together if this is called from the queue (only relevant if batch=True)
            preprocess: If False, will not run preprocessing of component data before running 'fn' (e.g. leaving it as a base64 string if this method is called with the `Image` component).
            postprocess: If False, will not run postprocessing of component data before returning 'fn' output to the browser.
            cancels: A list of other events to cancel when this listener is triggered. For example, setting cancels=[click_event] will cancel the click_event, where click_event is the return value of another components .click method. Functions that have not yet run (or generators that are iterating) will be cancelled, but functions that are currently running will be allowed to finish.
            every: Run this event 'every' number of seconds while the client connection is open. Interpreted in seconds.
            trigger_mode: If "once" (default for all events except `.change()`) would not allow any submissions while an event is pending. If set to "multiple", unlimited submissions are allowed while pending, and "always_last" (default for `.change()` event) would allow a second submission after the pending event is complete.
            js: Optional frontend js method to run before running 'fn'. Input arguments for js method are values of 'inputs' and 'outputs', return should be a list of values for output components.
            concurrency_limit: If set, this is the maximum number of this event that can be running simultaneously. Can be set to None to mean no concurrency_limit (any number of this event can be running simultaneously). Set to "default" to use the default concurrency limit (defined by the `default_concurrency_limit` parameter in `Blocks.queue()`, which itself is 1 by default).
            concurrency_id: If set, this is the id of the concurrency group. Events with the same concurrency_id will be limited by the lowest set concurrency_limit.
            show_api: whether to show this event in the "view API" page of the Gradio app, or in the ".view_api()" method of the Gradio clients. Unlike setting api_name to False, setting show_api to False will still allow downstream apps to use this event. If fn is None, show_api will automatically be set to False.
        """
        ...
    
    def delete_prompt(self,
        fn: Callable | None,
        inputs: Component | Sequence[Component] | set[Component] | None = None,
        outputs: Component | Sequence[Component] | None = None,
        api_name: str | None | Literal[False] = None,
        scroll_to_output: bool = False,
        show_progress: Literal["full", "minimal", "hidden"] = "full",
        queue: bool | None = None,
        batch: bool = False,
        max_batch_size: int = 4,
        preprocess: bool = True,
        postprocess: bool = True,
        cancels: dict[str, Any] | list[dict[str, Any]] | None = None,
        every: float | None = None,
        trigger_mode: Literal["once", "multiple", "always_last"] | None = None,
        js: str | None = None,
        concurrency_limit: int | None | Literal["default"] = "default",
        concurrency_id: str | None = None,
        show_api: bool = True) -> Dependency:
        """
        Parameters:
            fn: the function to call when this event is triggered. Often a machine learning model's prediction function. Each parameter of the function corresponds to one input component, and the function should return a single value or a tuple of values, with each element in the tuple corresponding to one output component.
            inputs: List of gradio.components to use as inputs. If the function takes no inputs, this should be an empty list.
            outputs: List of gradio.components to use as outputs. If the function returns no outputs, this should be an empty list.
            api_name: Defines how the endpoint appears in the API docs. Can be a string, None, or False. If False, the endpoint will not be exposed in the api docs. If set to None, the endpoint will be exposed in the api docs as an unnamed endpoint, although this behavior will be changed in Gradio 4.0. If set to a string, the endpoint will be exposed in the api docs with the given name.
            scroll_to_output: If True, will scroll to output component on completion
            show_progress: If True, will show progress animation while pending
            queue: If True, will place the request on the queue, if the queue has been enabled. If False, will not put this event on the queue, even if the queue has been enabled. If None, will use the queue setting of the gradio app.
            batch: If True, then the function should process a batch of inputs, meaning that it should accept a list of input values for each parameter. The lists should be of equal length (and be up to length `max_batch_size`). The function is then *required* to return a tuple of lists (even if there is only 1 output component), with each list in the tuple corresponding to one output component.
            max_batch_size: Maximum number of inputs to batch together if this is called from the queue (only relevant if batch=True)
            preprocess: If False, will not run preprocessing of component data before running 'fn' (e.g. leaving it as a base64 string if this method is called with the `Image` component).
            postprocess: If False, will not run postprocessing of component data before returning 'fn' output to the browser.
            cancels: A list of other events to cancel when this listener is triggered. For example, setting cancels=[click_event] will cancel the click_event, where click_event is the return value of another components .click method. Functions that have not yet run (or generators that are iterating) will be cancelled, but functions that are currently running will be allowed to finish.
            every: Run this event 'every' number of seconds while the client connection is open. Interpreted in seconds.
            trigger_mode: If "once" (default for all events except `.change()`) would not allow any submissions while an event is pending. If set to "multiple", unlimited submissions are allowed while pending, and "always_last" (default for `.change()` event) would allow a second submission after the pending event is complete.
            js: Optional frontend js method to run before running 'fn'. Input arguments for js method are values of 'inputs' and 'outputs', return should be a list of values for output components.
            concurrency_limit: If set, this is the maximum number of this event that can be running simultaneously. Can be set to None to mean no concurrency_limit (any number of this event can be running simultaneously). Set to "default" to use the default concurrency limit (defined by the `default_concurrency_limit` parameter in `Blocks.queue()`, which itself is 1 by default).
            concurrency_id: If set, this is the id of the concurrency group. Events with the same concurrency_id will be limited by the lowest set concurrency_limit.
            show_api: whether to show this event in the "view API" page of the Gradio app, or in the ".view_api()" method of the Gradio clients. Unlike setting api_name to False, setting show_api to False will still allow downstream apps to use this event. If fn is None, show_api will automatically be set to False.
        """
        ...
    
    def get_aiconfig(self,
        fn: Callable | None,
        inputs: Component | Sequence[Component] | set[Component] | None = None,
        outputs: Component | Sequence[Component] | None = None,
        api_name: str | None | Literal[False] = None,
        scroll_to_output: bool = False,
        show_progress: Literal["full", "minimal", "hidden"] = "full",
        queue: bool | None = None,
        batch: bool = False,
        max_batch_size: int = 4,
        preprocess: bool = True,
        postprocess: bool = True,
        cancels: dict[str, Any] | list[dict[str, Any]] | None = None,
        every: float | None = None,
        trigger_mode: Literal["once", "multiple", "always_last"] | None = None,
        js: str | None = None,
        concurrency_limit: int | None | Literal["default"] = "default",
        concurrency_id: str | None = None,
        show_api: bool = True) -> Dependency:
        """
        Parameters:
            fn: the function to call when this event is triggered. Often a machine learning model's prediction function. Each parameter of the function corresponds to one input component, and the function should return a single value or a tuple of values, with each element in the tuple corresponding to one output component.
            inputs: List of gradio.components to use as inputs. If the function takes no inputs, this should be an empty list.
            outputs: List of gradio.components to use as outputs. If the function returns no outputs, this should be an empty list.
            api_name: Defines how the endpoint appears in the API docs. Can be a string, None, or False. If False, the endpoint will not be exposed in the api docs. If set to None, the endpoint will be exposed in the api docs as an unnamed endpoint, although this behavior will be changed in Gradio 4.0. If set to a string, the endpoint will be exposed in the api docs with the given name.
            scroll_to_output: If True, will scroll to output component on completion
            show_progress: If True, will show progress animation while pending
            queue: If True, will place the request on the queue, if the queue has been enabled. If False, will not put this event on the queue, even if the queue has been enabled. If None, will use the queue setting of the gradio app.
            batch: If True, then the function should process a batch of inputs, meaning that it should accept a list of input values for each parameter. The lists should be of equal length (and be up to length `max_batch_size`). The function is then *required* to return a tuple of lists (even if there is only 1 output component), with each list in the tuple corresponding to one output component.
            max_batch_size: Maximum number of inputs to batch together if this is called from the queue (only relevant if batch=True)
            preprocess: If False, will not run preprocessing of component data before running 'fn' (e.g. leaving it as a base64 string if this method is called with the `Image` component).
            postprocess: If False, will not run postprocessing of component data before returning 'fn' output to the browser.
            cancels: A list of other events to cancel when this listener is triggered. For example, setting cancels=[click_event] will cancel the click_event, where click_event is the return value of another components .click method. Functions that have not yet run (or generators that are iterating) will be cancelled, but functions that are currently running will be allowed to finish.
            every: Run this event 'every' number of seconds while the client connection is open. Interpreted in seconds.
            trigger_mode: If "once" (default for all events except `.change()`) would not allow any submissions while an event is pending. If set to "multiple", unlimited submissions are allowed while pending, and "always_last" (default for `.change()` event) would allow a second submission after the pending event is complete.
            js: Optional frontend js method to run before running 'fn'. Input arguments for js method are values of 'inputs' and 'outputs', return should be a list of values for output components.
            concurrency_limit: If set, this is the maximum number of this event that can be running simultaneously. Can be set to None to mean no concurrency_limit (any number of this event can be running simultaneously). Set to "default" to use the default concurrency limit (defined by the `default_concurrency_limit` parameter in `Blocks.queue()`, which itself is 1 by default).
            concurrency_id: If set, this is the id of the concurrency group. Events with the same concurrency_id will be limited by the lowest set concurrency_limit.
            show_api: whether to show this event in the "view API" page of the Gradio app, or in the ".view_api()" method of the Gradio clients. Unlike setting api_name to False, setting show_api to False will still allow downstream apps to use this event. If fn is None, show_api will automatically be set to False.
        """
        ...
    
    def remove_session_id(self,
        fn: Callable | None,
        inputs: Component | Sequence[Component] | set[Component] | None = None,
        outputs: Component | Sequence[Component] | None = None,
        api_name: str | None | Literal[False] = None,
        scroll_to_output: bool = False,
        show_progress: Literal["full", "minimal", "hidden"] = "full",
        queue: bool | None = None,
        batch: bool = False,
        max_batch_size: int = 4,
        preprocess: bool = True,
        postprocess: bool = True,
        cancels: dict[str, Any] | list[dict[str, Any]] | None = None,
        every: float | None = None,
        trigger_mode: Literal["once", "multiple", "always_last"] | None = None,
        js: str | None = None,
        concurrency_limit: int | None | Literal["default"] = "default",
        concurrency_id: str | None = None,
        show_api: bool = True) -> Dependency:
        """
        Parameters:
            fn: the function to call when this event is triggered. Often a machine learning model's prediction function. Each parameter of the function corresponds to one input component, and the function should return a single value or a tuple of values, with each element in the tuple corresponding to one output component.
            inputs: List of gradio.components to use as inputs. If the function takes no inputs, this should be an empty list.
            outputs: List of gradio.components to use as outputs. If the function returns no outputs, this should be an empty list.
            api_name: Defines how the endpoint appears in the API docs. Can be a string, None, or False. If False, the endpoint will not be exposed in the api docs. If set to None, the endpoint will be exposed in the api docs as an unnamed endpoint, although this behavior will be changed in Gradio 4.0. If set to a string, the endpoint will be exposed in the api docs with the given name.
            scroll_to_output: If True, will scroll to output component on completion
            show_progress: If True, will show progress animation while pending
            queue: If True, will place the request on the queue, if the queue has been enabled. If False, will not put this event on the queue, even if the queue has been enabled. If None, will use the queue setting of the gradio app.
            batch: If True, then the function should process a batch of inputs, meaning that it should accept a list of input values for each parameter. The lists should be of equal length (and be up to length `max_batch_size`). The function is then *required* to return a tuple of lists (even if there is only 1 output component), with each list in the tuple corresponding to one output component.
            max_batch_size: Maximum number of inputs to batch together if this is called from the queue (only relevant if batch=True)
            preprocess: If False, will not run preprocessing of component data before running 'fn' (e.g. leaving it as a base64 string if this method is called with the `Image` component).
            postprocess: If False, will not run postprocessing of component data before returning 'fn' output to the browser.
            cancels: A list of other events to cancel when this listener is triggered. For example, setting cancels=[click_event] will cancel the click_event, where click_event is the return value of another components .click method. Functions that have not yet run (or generators that are iterating) will be cancelled, but functions that are currently running will be allowed to finish.
            every: Run this event 'every' number of seconds while the client connection is open. Interpreted in seconds.
            trigger_mode: If "once" (default for all events except `.change()`) would not allow any submissions while an event is pending. If set to "multiple", unlimited submissions are allowed while pending, and "always_last" (default for `.change()` event) would allow a second submission after the pending event is complete.
            js: Optional frontend js method to run before running 'fn'. Input arguments for js method are values of 'inputs' and 'outputs', return should be a list of values for output components.
            concurrency_limit: If set, this is the maximum number of this event that can be running simultaneously. Can be set to None to mean no concurrency_limit (any number of this event can be running simultaneously). Set to "default" to use the default concurrency limit (defined by the `default_concurrency_limit` parameter in `Blocks.queue()`, which itself is 1 by default).
            concurrency_id: If set, this is the id of the concurrency group. Events with the same concurrency_id will be limited by the lowest set concurrency_limit.
            show_api: whether to show this event in the "view API" page of the Gradio app, or in the ".view_api()" method of the Gradio clients. Unlike setting api_name to False, setting show_api to False will still allow downstream apps to use this event. If fn is None, show_api will automatically be set to False.
        """
        ...
    
    def run_prompt(self,
        fn: Callable | None,
        inputs: Component | Sequence[Component] | set[Component] | None = None,
        outputs: Component | Sequence[Component] | None = None,
        api_name: str | None | Literal[False] = None,
        scroll_to_output: bool = False,
        show_progress: Literal["full", "minimal", "hidden"] = "full",
        queue: bool | None = None,
        batch: bool = False,
        max_batch_size: int = 4,
        preprocess: bool = True,
        postprocess: bool = True,
        cancels: dict[str, Any] | list[dict[str, Any]] | None = None,
        every: float | None = None,
        trigger_mode: Literal["once", "multiple", "always_last"] | None = None,
        js: str | None = None,
        concurrency_limit: int | None | Literal["default"] = "default",
        concurrency_id: str | None = None,
        show_api: bool = True) -> Dependency:
        """
        Parameters:
            fn: the function to call when this event is triggered. Often a machine learning model's prediction function. Each parameter of the function corresponds to one input component, and the function should return a single value or a tuple of values, with each element in the tuple corresponding to one output component.
            inputs: List of gradio.components to use as inputs. If the function takes no inputs, this should be an empty list.
            outputs: List of gradio.components to use as outputs. If the function returns no outputs, this should be an empty list.
            api_name: Defines how the endpoint appears in the API docs. Can be a string, None, or False. If False, the endpoint will not be exposed in the api docs. If set to None, the endpoint will be exposed in the api docs as an unnamed endpoint, although this behavior will be changed in Gradio 4.0. If set to a string, the endpoint will be exposed in the api docs with the given name.
            scroll_to_output: If True, will scroll to output component on completion
            show_progress: If True, will show progress animation while pending
            queue: If True, will place the request on the queue, if the queue has been enabled. If False, will not put this event on the queue, even if the queue has been enabled. If None, will use the queue setting of the gradio app.
            batch: If True, then the function should process a batch of inputs, meaning that it should accept a list of input values for each parameter. The lists should be of equal length (and be up to length `max_batch_size`). The function is then *required* to return a tuple of lists (even if there is only 1 output component), with each list in the tuple corresponding to one output component.
            max_batch_size: Maximum number of inputs to batch together if this is called from the queue (only relevant if batch=True)
            preprocess: If False, will not run preprocessing of component data before running 'fn' (e.g. leaving it as a base64 string if this method is called with the `Image` component).
            postprocess: If False, will not run postprocessing of component data before returning 'fn' output to the browser.
            cancels: A list of other events to cancel when this listener is triggered. For example, setting cancels=[click_event] will cancel the click_event, where click_event is the return value of another components .click method. Functions that have not yet run (or generators that are iterating) will be cancelled, but functions that are currently running will be allowed to finish.
            every: Run this event 'every' number of seconds while the client connection is open. Interpreted in seconds.
            trigger_mode: If "once" (default for all events except `.change()`) would not allow any submissions while an event is pending. If set to "multiple", unlimited submissions are allowed while pending, and "always_last" (default for `.change()` event) would allow a second submission after the pending event is complete.
            js: Optional frontend js method to run before running 'fn'. Input arguments for js method are values of 'inputs' and 'outputs', return should be a list of values for output components.
            concurrency_limit: If set, this is the maximum number of this event that can be running simultaneously. Can be set to None to mean no concurrency_limit (any number of this event can be running simultaneously). Set to "default" to use the default concurrency limit (defined by the `default_concurrency_limit` parameter in `Blocks.queue()`, which itself is 1 by default).
            concurrency_id: If set, this is the id of the concurrency group. Events with the same concurrency_id will be limited by the lowest set concurrency_limit.
            show_api: whether to show this event in the "view API" page of the Gradio app, or in the ".view_api()" method of the Gradio clients. Unlike setting api_name to False, setting show_api to False will still allow downstream apps to use this event. If fn is None, show_api will automatically be set to False.
        """
        ...
    
    def share_config(self,
        fn: Callable | None,
        inputs: Component | Sequence[Component] | set[Component] | None = None,
        outputs: Component | Sequence[Component] | None = None,
        api_name: str | None | Literal[False] = None,
        scroll_to_output: bool = False,
        show_progress: Literal["full", "minimal", "hidden"] = "full",
        queue: bool | None = None,
        batch: bool = False,
        max_batch_size: int = 4,
        preprocess: bool = True,
        postprocess: bool = True,
        cancels: dict[str, Any] | list[dict[str, Any]] | None = None,
        every: float | None = None,
        trigger_mode: Literal["once", "multiple", "always_last"] | None = None,
        js: str | None = None,
        concurrency_limit: int | None | Literal["default"] = "default",
        concurrency_id: str | None = None,
        show_api: bool = True) -> Dependency:
        """
        Parameters:
            fn: the function to call when this event is triggered. Often a machine learning model's prediction function. Each parameter of the function corresponds to one input component, and the function should return a single value or a tuple of values, with each element in the tuple corresponding to one output component.
            inputs: List of gradio.components to use as inputs. If the function takes no inputs, this should be an empty list.
            outputs: List of gradio.components to use as outputs. If the function returns no outputs, this should be an empty list.
            api_name: Defines how the endpoint appears in the API docs. Can be a string, None, or False. If False, the endpoint will not be exposed in the api docs. If set to None, the endpoint will be exposed in the api docs as an unnamed endpoint, although this behavior will be changed in Gradio 4.0. If set to a string, the endpoint will be exposed in the api docs with the given name.
            scroll_to_output: If True, will scroll to output component on completion
            show_progress: If True, will show progress animation while pending
            queue: If True, will place the request on the queue, if the queue has been enabled. If False, will not put this event on the queue, even if the queue has been enabled. If None, will use the queue setting of the gradio app.
            batch: If True, then the function should process a batch of inputs, meaning that it should accept a list of input values for each parameter. The lists should be of equal length (and be up to length `max_batch_size`). The function is then *required* to return a tuple of lists (even if there is only 1 output component), with each list in the tuple corresponding to one output component.
            max_batch_size: Maximum number of inputs to batch together if this is called from the queue (only relevant if batch=True)
            preprocess: If False, will not run preprocessing of component data before running 'fn' (e.g. leaving it as a base64 string if this method is called with the `Image` component).
            postprocess: If False, will not run postprocessing of component data before returning 'fn' output to the browser.
            cancels: A list of other events to cancel when this listener is triggered. For example, setting cancels=[click_event] will cancel the click_event, where click_event is the return value of another components .click method. Functions that have not yet run (or generators that are iterating) will be cancelled, but functions that are currently running will be allowed to finish.
            every: Run this event 'every' number of seconds while the client connection is open. Interpreted in seconds.
            trigger_mode: If "once" (default for all events except `.change()`) would not allow any submissions while an event is pending. If set to "multiple", unlimited submissions are allowed while pending, and "always_last" (default for `.change()` event) would allow a second submission after the pending event is complete.
            js: Optional frontend js method to run before running 'fn'. Input arguments for js method are values of 'inputs' and 'outputs', return should be a list of values for output components.
            concurrency_limit: If set, this is the maximum number of this event that can be running simultaneously. Can be set to None to mean no concurrency_limit (any number of this event can be running simultaneously). Set to "default" to use the default concurrency limit (defined by the `default_concurrency_limit` parameter in `Blocks.queue()`, which itself is 1 by default).
            concurrency_id: If set, this is the id of the concurrency group. Events with the same concurrency_id will be limited by the lowest set concurrency_limit.
            show_api: whether to show this event in the "view API" page of the Gradio app, or in the ".view_api()" method of the Gradio clients. Unlike setting api_name to False, setting show_api to False will still allow downstream apps to use this event. If fn is None, show_api will automatically be set to False.
        """
        ...
    
    def set_config_description(self,
        fn: Callable | None,
        inputs: Component | Sequence[Component] | set[Component] | None = None,
        outputs: Component | Sequence[Component] | None = None,
        api_name: str | None | Literal[False] = None,
        scroll_to_output: bool = False,
        show_progress: Literal["full", "minimal", "hidden"] = "full",
        queue: bool | None = None,
        batch: bool = False,
        max_batch_size: int = 4,
        preprocess: bool = True,
        postprocess: bool = True,
        cancels: dict[str, Any] | list[dict[str, Any]] | None = None,
        every: float | None = None,
        trigger_mode: Literal["once", "multiple", "always_last"] | None = None,
        js: str | None = None,
        concurrency_limit: int | None | Literal["default"] = "default",
        concurrency_id: str | None = None,
        show_api: bool = True) -> Dependency:
        """
        Parameters:
            fn: the function to call when this event is triggered. Often a machine learning model's prediction function. Each parameter of the function corresponds to one input component, and the function should return a single value or a tuple of values, with each element in the tuple corresponding to one output component.
            inputs: List of gradio.components to use as inputs. If the function takes no inputs, this should be an empty list.
            outputs: List of gradio.components to use as outputs. If the function returns no outputs, this should be an empty list.
            api_name: Defines how the endpoint appears in the API docs. Can be a string, None, or False. If False, the endpoint will not be exposed in the api docs. If set to None, the endpoint will be exposed in the api docs as an unnamed endpoint, although this behavior will be changed in Gradio 4.0. If set to a string, the endpoint will be exposed in the api docs with the given name.
            scroll_to_output: If True, will scroll to output component on completion
            show_progress: If True, will show progress animation while pending
            queue: If True, will place the request on the queue, if the queue has been enabled. If False, will not put this event on the queue, even if the queue has been enabled. If None, will use the queue setting of the gradio app.
            batch: If True, then the function should process a batch of inputs, meaning that it should accept a list of input values for each parameter. The lists should be of equal length (and be up to length `max_batch_size`). The function is then *required* to return a tuple of lists (even if there is only 1 output component), with each list in the tuple corresponding to one output component.
            max_batch_size: Maximum number of inputs to batch together if this is called from the queue (only relevant if batch=True)
            preprocess: If False, will not run preprocessing of component data before running 'fn' (e.g. leaving it as a base64 string if this method is called with the `Image` component).
            postprocess: If False, will not run postprocessing of component data before returning 'fn' output to the browser.
            cancels: A list of other events to cancel when this listener is triggered. For example, setting cancels=[click_event] will cancel the click_event, where click_event is the return value of another components .click method. Functions that have not yet run (or generators that are iterating) will be cancelled, but functions that are currently running will be allowed to finish.
            every: Run this event 'every' number of seconds while the client connection is open. Interpreted in seconds.
            trigger_mode: If "once" (default for all events except `.change()`) would not allow any submissions while an event is pending. If set to "multiple", unlimited submissions are allowed while pending, and "always_last" (default for `.change()` event) would allow a second submission after the pending event is complete.
            js: Optional frontend js method to run before running 'fn'. Input arguments for js method are values of 'inputs' and 'outputs', return should be a list of values for output components.
            concurrency_limit: If set, this is the maximum number of this event that can be running simultaneously. Can be set to None to mean no concurrency_limit (any number of this event can be running simultaneously). Set to "default" to use the default concurrency limit (defined by the `default_concurrency_limit` parameter in `Blocks.queue()`, which itself is 1 by default).
            concurrency_id: If set, this is the id of the concurrency group. Events with the same concurrency_id will be limited by the lowest set concurrency_limit.
            show_api: whether to show this event in the "view API" page of the Gradio app, or in the ".view_api()" method of the Gradio clients. Unlike setting api_name to False, setting show_api to False will still allow downstream apps to use this event. If fn is None, show_api will automatically be set to False.
        """
        ...
    
    def set_config_name(self,
        fn: Callable | None,
        inputs: Component | Sequence[Component] | set[Component] | None = None,
        outputs: Component | Sequence[Component] | None = None,
        api_name: str | None | Literal[False] = None,
        scroll_to_output: bool = False,
        show_progress: Literal["full", "minimal", "hidden"] = "full",
        queue: bool | None = None,
        batch: bool = False,
        max_batch_size: int = 4,
        preprocess: bool = True,
        postprocess: bool = True,
        cancels: dict[str, Any] | list[dict[str, Any]] | None = None,
        every: float | None = None,
        trigger_mode: Literal["once", "multiple", "always_last"] | None = None,
        js: str | None = None,
        concurrency_limit: int | None | Literal["default"] = "default",
        concurrency_id: str | None = None,
        show_api: bool = True) -> Dependency:
        """
        Parameters:
            fn: the function to call when this event is triggered. Often a machine learning model's prediction function. Each parameter of the function corresponds to one input component, and the function should return a single value or a tuple of values, with each element in the tuple corresponding to one output component.
            inputs: List of gradio.components to use as inputs. If the function takes no inputs, this should be an empty list.
            outputs: List of gradio.components to use as outputs. If the function returns no outputs, this should be an empty list.
            api_name: Defines how the endpoint appears in the API docs. Can be a string, None, or False. If False, the endpoint will not be exposed in the api docs. If set to None, the endpoint will be exposed in the api docs as an unnamed endpoint, although this behavior will be changed in Gradio 4.0. If set to a string, the endpoint will be exposed in the api docs with the given name.
            scroll_to_output: If True, will scroll to output component on completion
            show_progress: If True, will show progress animation while pending
            queue: If True, will place the request on the queue, if the queue has been enabled. If False, will not put this event on the queue, even if the queue has been enabled. If None, will use the queue setting of the gradio app.
            batch: If True, then the function should process a batch of inputs, meaning that it should accept a list of input values for each parameter. The lists should be of equal length (and be up to length `max_batch_size`). The function is then *required* to return a tuple of lists (even if there is only 1 output component), with each list in the tuple corresponding to one output component.
            max_batch_size: Maximum number of inputs to batch together if this is called from the queue (only relevant if batch=True)
            preprocess: If False, will not run preprocessing of component data before running 'fn' (e.g. leaving it as a base64 string if this method is called with the `Image` component).
            postprocess: If False, will not run postprocessing of component data before returning 'fn' output to the browser.
            cancels: A list of other events to cancel when this listener is triggered. For example, setting cancels=[click_event] will cancel the click_event, where click_event is the return value of another components .click method. Functions that have not yet run (or generators that are iterating) will be cancelled, but functions that are currently running will be allowed to finish.
            every: Run this event 'every' number of seconds while the client connection is open. Interpreted in seconds.
            trigger_mode: If "once" (default for all events except `.change()`) would not allow any submissions while an event is pending. If set to "multiple", unlimited submissions are allowed while pending, and "always_last" (default for `.change()` event) would allow a second submission after the pending event is complete.
            js: Optional frontend js method to run before running 'fn'. Input arguments for js method are values of 'inputs' and 'outputs', return should be a list of values for output components.
            concurrency_limit: If set, this is the maximum number of this event that can be running simultaneously. Can be set to None to mean no concurrency_limit (any number of this event can be running simultaneously). Set to "default" to use the default concurrency limit (defined by the `default_concurrency_limit` parameter in `Blocks.queue()`, which itself is 1 by default).
            concurrency_id: If set, this is the id of the concurrency group. Events with the same concurrency_id will be limited by the lowest set concurrency_limit.
            show_api: whether to show this event in the "view API" page of the Gradio app, or in the ".view_api()" method of the Gradio clients. Unlike setting api_name to False, setting show_api to False will still allow downstream apps to use this event. If fn is None, show_api will automatically be set to False.
        """
        ...
    
    def set_parameters(self,
        fn: Callable | None,
        inputs: Component | Sequence[Component] | set[Component] | None = None,
        outputs: Component | Sequence[Component] | None = None,
        api_name: str | None | Literal[False] = None,
        scroll_to_output: bool = False,
        show_progress: Literal["full", "minimal", "hidden"] = "full",
        queue: bool | None = None,
        batch: bool = False,
        max_batch_size: int = 4,
        preprocess: bool = True,
        postprocess: bool = True,
        cancels: dict[str, Any] | list[dict[str, Any]] | None = None,
        every: float | None = None,
        trigger_mode: Literal["once", "multiple", "always_last"] | None = None,
        js: str | None = None,
        concurrency_limit: int | None | Literal["default"] = "default",
        concurrency_id: str | None = None,
        show_api: bool = True) -> Dependency:
        """
        Parameters:
            fn: the function to call when this event is triggered. Often a machine learning model's prediction function. Each parameter of the function corresponds to one input component, and the function should return a single value or a tuple of values, with each element in the tuple corresponding to one output component.
            inputs: List of gradio.components to use as inputs. If the function takes no inputs, this should be an empty list.
            outputs: List of gradio.components to use as outputs. If the function returns no outputs, this should be an empty list.
            api_name: Defines how the endpoint appears in the API docs. Can be a string, None, or False. If False, the endpoint will not be exposed in the api docs. If set to None, the endpoint will be exposed in the api docs as an unnamed endpoint, although this behavior will be changed in Gradio 4.0. If set to a string, the endpoint will be exposed in the api docs with the given name.
            scroll_to_output: If True, will scroll to output component on completion
            show_progress: If True, will show progress animation while pending
            queue: If True, will place the request on the queue, if the queue has been enabled. If False, will not put this event on the queue, even if the queue has been enabled. If None, will use the queue setting of the gradio app.
            batch: If True, then the function should process a batch of inputs, meaning that it should accept a list of input values for each parameter. The lists should be of equal length (and be up to length `max_batch_size`). The function is then *required* to return a tuple of lists (even if there is only 1 output component), with each list in the tuple corresponding to one output component.
            max_batch_size: Maximum number of inputs to batch together if this is called from the queue (only relevant if batch=True)
            preprocess: If False, will not run preprocessing of component data before running 'fn' (e.g. leaving it as a base64 string if this method is called with the `Image` component).
            postprocess: If False, will not run postprocessing of component data before returning 'fn' output to the browser.
            cancels: A list of other events to cancel when this listener is triggered. For example, setting cancels=[click_event] will cancel the click_event, where click_event is the return value of another components .click method. Functions that have not yet run (or generators that are iterating) will be cancelled, but functions that are currently running will be allowed to finish.
            every: Run this event 'every' number of seconds while the client connection is open. Interpreted in seconds.
            trigger_mode: If "once" (default for all events except `.change()`) would not allow any submissions while an event is pending. If set to "multiple", unlimited submissions are allowed while pending, and "always_last" (default for `.change()` event) would allow a second submission after the pending event is complete.
            js: Optional frontend js method to run before running 'fn'. Input arguments for js method are values of 'inputs' and 'outputs', return should be a list of values for output components.
            concurrency_limit: If set, this is the maximum number of this event that can be running simultaneously. Can be set to None to mean no concurrency_limit (any number of this event can be running simultaneously). Set to "default" to use the default concurrency limit (defined by the `default_concurrency_limit` parameter in `Blocks.queue()`, which itself is 1 by default).
            concurrency_id: If set, this is the id of the concurrency group. Events with the same concurrency_id will be limited by the lowest set concurrency_limit.
            show_api: whether to show this event in the "view API" page of the Gradio app, or in the ".view_api()" method of the Gradio clients. Unlike setting api_name to False, setting show_api to False will still allow downstream apps to use this event. If fn is None, show_api will automatically be set to False.
        """
        ...
    
    def update_model(self,
        fn: Callable | None,
        inputs: Component | Sequence[Component] | set[Component] | None = None,
        outputs: Component | Sequence[Component] | None = None,
        api_name: str | None | Literal[False] = None,
        scroll_to_output: bool = False,
        show_progress: Literal["full", "minimal", "hidden"] = "full",
        queue: bool | None = None,
        batch: bool = False,
        max_batch_size: int = 4,
        preprocess: bool = True,
        postprocess: bool = True,
        cancels: dict[str, Any] | list[dict[str, Any]] | None = None,
        every: float | None = None,
        trigger_mode: Literal["once", "multiple", "always_last"] | None = None,
        js: str | None = None,
        concurrency_limit: int | None | Literal["default"] = "default",
        concurrency_id: str | None = None,
        show_api: bool = True) -> Dependency:
        """
        Parameters:
            fn: the function to call when this event is triggered. Often a machine learning model's prediction function. Each parameter of the function corresponds to one input component, and the function should return a single value or a tuple of values, with each element in the tuple corresponding to one output component.
            inputs: List of gradio.components to use as inputs. If the function takes no inputs, this should be an empty list.
            outputs: List of gradio.components to use as outputs. If the function returns no outputs, this should be an empty list.
            api_name: Defines how the endpoint appears in the API docs. Can be a string, None, or False. If False, the endpoint will not be exposed in the api docs. If set to None, the endpoint will be exposed in the api docs as an unnamed endpoint, although this behavior will be changed in Gradio 4.0. If set to a string, the endpoint will be exposed in the api docs with the given name.
            scroll_to_output: If True, will scroll to output component on completion
            show_progress: If True, will show progress animation while pending
            queue: If True, will place the request on the queue, if the queue has been enabled. If False, will not put this event on the queue, even if the queue has been enabled. If None, will use the queue setting of the gradio app.
            batch: If True, then the function should process a batch of inputs, meaning that it should accept a list of input values for each parameter. The lists should be of equal length (and be up to length `max_batch_size`). The function is then *required* to return a tuple of lists (even if there is only 1 output component), with each list in the tuple corresponding to one output component.
            max_batch_size: Maximum number of inputs to batch together if this is called from the queue (only relevant if batch=True)
            preprocess: If False, will not run preprocessing of component data before running 'fn' (e.g. leaving it as a base64 string if this method is called with the `Image` component).
            postprocess: If False, will not run postprocessing of component data before returning 'fn' output to the browser.
            cancels: A list of other events to cancel when this listener is triggered. For example, setting cancels=[click_event] will cancel the click_event, where click_event is the return value of another components .click method. Functions that have not yet run (or generators that are iterating) will be cancelled, but functions that are currently running will be allowed to finish.
            every: Run this event 'every' number of seconds while the client connection is open. Interpreted in seconds.
            trigger_mode: If "once" (default for all events except `.change()`) would not allow any submissions while an event is pending. If set to "multiple", unlimited submissions are allowed while pending, and "always_last" (default for `.change()` event) would allow a second submission after the pending event is complete.
            js: Optional frontend js method to run before running 'fn'. Input arguments for js method are values of 'inputs' and 'outputs', return should be a list of values for output components.
            concurrency_limit: If set, this is the maximum number of this event that can be running simultaneously. Can be set to None to mean no concurrency_limit (any number of this event can be running simultaneously). Set to "default" to use the default concurrency limit (defined by the `default_concurrency_limit` parameter in `Blocks.queue()`, which itself is 1 by default).
            concurrency_id: If set, this is the id of the concurrency group. Events with the same concurrency_id will be limited by the lowest set concurrency_limit.
            show_api: whether to show this event in the "view API" page of the Gradio app, or in the ".view_api()" method of the Gradio clients. Unlike setting api_name to False, setting show_api to False will still allow downstream apps to use this event. If fn is None, show_api will automatically be set to False.
        """
        ...
    
    def update_prompt(self,
        fn: Callable | None,
        inputs: Component | Sequence[Component] | set[Component] | None = None,
        outputs: Component | Sequence[Component] | None = None,
        api_name: str | None | Literal[False] = None,
        scroll_to_output: bool = False,
        show_progress: Literal["full", "minimal", "hidden"] = "full",
        queue: bool | None = None,
        batch: bool = False,
        max_batch_size: int = 4,
        preprocess: bool = True,
        postprocess: bool = True,
        cancels: dict[str, Any] | list[dict[str, Any]] | None = None,
        every: float | None = None,
        trigger_mode: Literal["once", "multiple", "always_last"] | None = None,
        js: str | None = None,
        concurrency_limit: int | None | Literal["default"] = "default",
        concurrency_id: str | None = None,
        show_api: bool = True) -> Dependency:
        """
        Parameters:
            fn: the function to call when this event is triggered. Often a machine learning model's prediction function. Each parameter of the function corresponds to one input component, and the function should return a single value or a tuple of values, with each element in the tuple corresponding to one output component.
            inputs: List of gradio.components to use as inputs. If the function takes no inputs, this should be an empty list.
            outputs: List of gradio.components to use as outputs. If the function returns no outputs, this should be an empty list.
            api_name: Defines how the endpoint appears in the API docs. Can be a string, None, or False. If False, the endpoint will not be exposed in the api docs. If set to None, the endpoint will be exposed in the api docs as an unnamed endpoint, although this behavior will be changed in Gradio 4.0. If set to a string, the endpoint will be exposed in the api docs with the given name.
            scroll_to_output: If True, will scroll to output component on completion
            show_progress: If True, will show progress animation while pending
            queue: If True, will place the request on the queue, if the queue has been enabled. If False, will not put this event on the queue, even if the queue has been enabled. If None, will use the queue setting of the gradio app.
            batch: If True, then the function should process a batch of inputs, meaning that it should accept a list of input values for each parameter. The lists should be of equal length (and be up to length `max_batch_size`). The function is then *required* to return a tuple of lists (even if there is only 1 output component), with each list in the tuple corresponding to one output component.
            max_batch_size: Maximum number of inputs to batch together if this is called from the queue (only relevant if batch=True)
            preprocess: If False, will not run preprocessing of component data before running 'fn' (e.g. leaving it as a base64 string if this method is called with the `Image` component).
            postprocess: If False, will not run postprocessing of component data before returning 'fn' output to the browser.
            cancels: A list of other events to cancel when this listener is triggered. For example, setting cancels=[click_event] will cancel the click_event, where click_event is the return value of another components .click method. Functions that have not yet run (or generators that are iterating) will be cancelled, but functions that are currently running will be allowed to finish.
            every: Run this event 'every' number of seconds while the client connection is open. Interpreted in seconds.
            trigger_mode: If "once" (default for all events except `.change()`) would not allow any submissions while an event is pending. If set to "multiple", unlimited submissions are allowed while pending, and "always_last" (default for `.change()` event) would allow a second submission after the pending event is complete.
            js: Optional frontend js method to run before running 'fn'. Input arguments for js method are values of 'inputs' and 'outputs', return should be a list of values for output components.
            concurrency_limit: If set, this is the maximum number of this event that can be running simultaneously. Can be set to None to mean no concurrency_limit (any number of this event can be running simultaneously). Set to "default" to use the default concurrency limit (defined by the `default_concurrency_limit` parameter in `Blocks.queue()`, which itself is 1 by default).
            concurrency_id: If set, this is the id of the concurrency group. Events with the same concurrency_id will be limited by the lowest set concurrency_limit.
            show_api: whether to show this event in the "view API" page of the Gradio app, or in the ".view_api()" method of the Gradio clients. Unlike setting api_name to False, setting show_api to False will still allow downstream apps to use this event. If fn is None, show_api will automatically be set to False.
        """
        ...